# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Utilities for preparing inputs."""

import attrs
import numpy as np
from numpy.typing import NDArray

__all__ = (
    "PositiveDefiniteError",
    "UnitConfig",
    "block_average",
    "label_unit",
    "mixture_stats",
    "robust_dot",
    "robust_posinv",
    "split",
)


@attrs.define
class UnitConfig:
    """Unit configuration for functions that print or plot values.

    This class never influences numerical values in STACIE's computations,
    such as attributes of a result object or other variables.
    It only affects printed or plotted values.

    The ``acint_unit``, ``freq_unit``, and ``time_unit`` attributes should be set as follows:

    - The values of variables in STACIE (and your scripts using STACIE) are always
      in "internal units".
    - The ``*_unit`` attributes are assumed to have the value of a "display unit" in
      the same internal units.

    For example, if your internal time unit is 1 ps and you want to display times
    in ns, set ``time_unit=1000.0``, because your display unit (1 ns) is 1000 internal units (1 ps).

    To make these conventions easy to follow (and to avoid unit hell in general),
    it is recommended to pick consistent internal units for your system.
    For example, use atomic units or SI units throughout your code:

    - As soon as you load data from a file, immediately convert it to internal units.
    - Only just before printing or plotting, convert to display units,
      which is also how this class is used in STACIE.

    For example, when all variables are in SI base units and you want to display time in ns,
    frequency in THz, and autocorrelation integrals in cm^2/s, then create a ``UnitConfig``
    as follows:

    .. code-block:: python

        units = UnitConfig(
            time_unit=1e-9,
            time_unit_str="ns",
            freq_unit=1e12,
            freq_unit_str="THz",
            acint_unit=1e-4,
            acint_unit_str="cm^2/s",
        )
    """

    acint_symbol: str = attrs.field(default=r"\mathcal{I}", kw_only=True)
    """The symbol used for the autocorrelation integral."""

    acint_unit_str: str = attrs.field(default="", kw_only=True)
    """The text used for the autocorrelation integral unit."""

    acint_unit: float = attrs.field(default=1.0, kw_only=True)
    """The unit of an autocorrelation integral."""

    acint_fmt: str = attrs.field(default=".2e", kw_only=True)
    """The format string for an autocorrelation integral."""

    freq_unit_str: str = attrs.field(default="", kw_only=True)
    """The text used for the frequency unit."""

    freq_unit: float = attrs.field(default=1.0, kw_only=True)
    """The unit of frequency."""

    freq_fmt: str = attrs.field(default=".2e", kw_only=True)
    """The format string for a frequency."""

    time_unit_str: str = attrs.field(default="", kw_only=True)
    """The text used for the time unit."""

    time_unit: float = attrs.field(default=1.0, kw_only=True)
    """The unit of time."""

    time_fmt: str = attrs.field(default=".2e", kw_only=True)
    """The format string for a time value."""

    clevel: float = attrs.field(default=0.95, kw_only=True)
    """The confidence level used to plot confidence intervals."""

    @property
    def clb(self) -> float:
        """The confidence lower bound used to plot confidence intervals."""
        return (1 - self.clevel) / 2

    @property
    def cub(self) -> float:
        """The confidence upper bound used to plot confidence intervals."""
        return (1 + self.clevel) / 2


def label_unit(label: str, unit_str: str | None) -> str:
    """Format a label with the unit string as `label [unit]`.

    When the unit is ``""`` or ``None``, the unit is omitted.

    Parameters
    ----------
    label
        The label text.
    unit_str
        The unit string.
    """
    if unit_str in ("", None):
        return label
    return f"{label} [{unit_str}]"


def split(sequences: NDArray[float], nsplit: int) -> NDArray:
    """Split input sequences into shorter parts of equal length.

    This reduces the resolution of the frequency axis of the spectrum,
    which may be useful when the sequence length is much longer than the exponential
    autocorrelation time.

    Parameters
    ----------
    sequences
        Input sequence(s) to be split, with shape ``(nseq, nstep)``.
        A single sequence with shape ``(nstep, )`` is also accepted.
    nsplit
        The number of splits.

    Returns
    -------
    split_sequences
        Splitted sequences, with shape ``(nseq * nsplit, nstep // nsplit)``.
    """
    sequences = np.asarray(sequences)
    if sequences.ndim == 1:
        sequences.shape = (1, -1)
    if not isinstance(nsplit, int) or nsplit <= 0 or nsplit > sequences.shape[-1] / 2:
        raise ValueError("nsplit must be a positive integer smaller than half the sequence length.")
    length = sequences.shape[1] // nsplit
    return sequences[:, : length * nsplit].reshape(-1, length)


def block_average(sequences: NDArray[float], size: int) -> NDArray:
    r"""Reduce input sequences by taking block averages.

    This reduces the maximum frequency of the frequency axis of the spectrum,
    which may be useful when the time step is much shorter than the exponential
    autocorrelation time.

    A time step :math:`h = \tau_\text{exp} / (20 \pi)` (after taking block averages)
    is recommended, not larger.

    Parameters
    ----------
    sequences
        Input sequence(s) to be block averaged, with shape ``(*data_shape, nstep)``,
        where ``data_shape`` represents any number of leading dimensions.
        A single sequence with shape ``(nstep, )`` is also accepted.
    size
        The block size

    Returns
    -------
    blav_sequences
        Sequences of block averages, with shape ``(*data_shape, nstep // size)``.
        If needed a few trailing elements of ``sequences`` are discarded
        to make ``nstep`` divisible by ``size``.
    """
    sequences = np.asarray(sequences)
    data_shape = sequences.shape[:-1]
    nblock = sequences.shape[-1] // size
    return sequences[..., : nblock * size].reshape(*data_shape, nblock, size).mean(axis=-1)


class PositiveDefiniteError(ValueError):
    """Raised when a matrix is not positive definite."""


def robust_posinv(matrix: NDArray[float]) -> tuple[NDArray, NDArray, NDArray, NDArray]:
    """Compute the eigenvalues, eigenvectors and inverse of a positive definite symmetric matrix.

    This function is a robust version of ``numpy.linalg.eigh`` and ``numpy.linalg.inv``
    that can handle large variations in order of magnitude of the diagonal elements.
    If the matrix is not positive definite, a ``ValueError`` is raised.

    Parameters
    ----------
    matrix
        Input matrix to be diagonalized.

    Returns
    -------
    scales
        The scales used to precondition the matrix.
    evals
        The eigenvalues of the preconditioned matrix.
    evecs
        The eigenvectors of the preconditioned matrix.
    inverse
        The inverse of the original.
    """
    matrix = np.asarray(matrix)
    if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
        raise PositiveDefiniteError("matrix is not square.")
    if not np.isfinite(matrix).all():
        raise PositiveDefiniteError("matrix contains NaN or inf.")
    matrix = 0.5 * (matrix + matrix.T)
    diag = np.diag(matrix)
    if diag.min() <= 0:
        raise PositiveDefiniteError(f"matrix has nonpositive diagonal elements: {diag=}.")
    scales = np.sqrt(np.diag(matrix))
    scaled_matrix = (matrix / scales[:, None]) / scales
    evals, evecs = np.linalg.eigh(scaled_matrix)
    if evals.min() <= 0:
        raise PositiveDefiniteError(f"matrix has non-positive eigenvalues: {evals=}")
    # Construct matrix square root of inverse first, to guarantee that the result is symmetric.
    half = evecs / np.sqrt(evals)
    scaled_inverse = np.dot(half, half.T)
    inverse = (scaled_inverse / scales[:, None]) / scales
    return scales, evals, evecs, inverse


def robust_dot(scales, evals, evecs, other):
    """Compute the dot product of a robustly diagonalized matrix with another matrix.

    - The first three arguments are the output of ``robust_posinv``.
    - To multiply with the inverse, just use element-wise inversion of ``scales`` and ``evals``.

    Parameters
    ----------
    scales
        The scales used to precondition the matrix.
    evals
        The eigenvalues of the preconditioned matrix.
    evecs
        The eigenvectors of the preconditioned matrix.
    other
        The other matrix to be multiplied. 1D or 2D arrays are accepted.

    Returns
    -------
    result
        The result of the dot product.
    """
    if other.ndim == 2:
        scales = scales[:, None]
        evals = evals[:, None]
    return np.dot(evecs, np.dot(evecs.T, other * scales) * evals) * scales


def mixture_stats(means: NDArray[float], covars: NDArray[float], weights: NDArray[float]):
    """Compute the statistics of the (Gaussian) mixture distribution.

    Parameters
    ----------
    means
        The means of the mixture components.
        Weighted averages are taken over the first index.
        Shape is ``(ncomp, nfeature)`` or ``(ncomp,)``.
        If the shape is ``(ncomp,)``, the means are interpreted as scalars.
        If the shape is ``(ncomp, nfeature)``, the means are interpreted as vectors.
    covars
        The covariances of the mixture components.
        If the shape matches that of the ``means`` argument,
        this array is interpreted as a diagonal covariance matrix.
        If the shape is ``(ncomp, nfeature, nfeature)``,
        this array is interpreted as full covariance matrices.
    weights
        The weights of the mixture components.
        Shape is ``(ncomp,)``.
        The weights are normalized to sum to 1.

    Returns
    -------
    mean
        The mean of the mixture distribution.
        Shape is ``(nfeature,)``.
    covar
        If the input covariance matrix is diagonal, the output covariance matrix
        is also diagonal and has shape ``(nfeature,)``.
        If the input covariance matrix is full, the output covariance matrix
        is also full and has shape ``(nfeature, nfeature)``.
    """
    means = np.asarray(means)
    covars = np.asarray(covars)
    weights = np.asarray(weights)
    if means.ndim < 1:
        raise ValueError("means must be at least a 1D array.")
    unpack = False
    if means.ndim == 1:
        means = means.reshape(-1, 1)
        unpack = True
    ncomp, nfeature = means.shape
    if covars.ndim < 1:
        raise ValueError("covars must be at least a 1D array.")
    if covars.ndim == 1:
        covars = covars.reshape(-1, 1)
    if weights.shape != (ncomp,):
        raise ValueError("weights must be a 1D vector with ncomp elements.")
    weights = weights / weights.sum()
    mean = np.dot(weights, means)
    deltas = means - mean
    if covars.shape == means.shape:
        covar = np.dot(weights, covars)
        covar += np.einsum("i,ij,ij->j", weights, deltas, deltas)
    elif covars.shape == (ncomp, nfeature, nfeature):
        covar = np.einsum("i,ijk->jk", weights, covars)
        covar += np.einsum("i,ij,ik->jk", weights, deltas, deltas)
    else:
        raise ValueError(
            f"Unsupported shape for covars: {covars.shape}. "
            f"Expected ({ncomp},), {means.shape}, or ({ncomp}, {nfeature}, {nfeature})."
        )
    if unpack:
        mean = mean[0]
        covar = covar[0]
    return mean, covar



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Regression tests for typical STACIE workflows.

Note that the test cases here are not meant as examples of properly analyzed spectra.
(Some of the fits are really poor.)
These test cases explore a few corner cases and verify that STACIE behaves as expected.
"""

from collections.abc import Callable
from pathlib import Path

import numpy as np
import pytest

from stacie.cutoff import CV2LCriterion
from stacie.estimate import Result, estimate_acint, summarize_results
from stacie.model import ExpPolyModel, PadeModel, SpectrumModel
from stacie.plot import plot_results
from stacie.spectrum import Spectrum

# Combinations of test spectra and suitable models, with a manual cutoff frequency.
CASES = [
    ("white", ExpPolyModel([0]), 0.1),
    ("white", ExpPolyModel([0, 1, 2]), 0.1),
    ("broad", ExpPolyModel([0]), 0.01),
    ("broad", ExpPolyModel([0, 1, 2]), 0.1),
    ("pure", PadeModel([0, 2], [2]), 0.1),
    ("pure", ExpPolyModel([0, 2]), 0.02),
    ("double", PadeModel([0, 2], [2]), 0.1),
    ("double", ExpPolyModel([0, 2]), 0.05),
    ("double", PadeModel([0, 2], [2]), 0.05),
]
CRITERIA = [CV2LCriterion()]


def output_test_result(prefix: str, res: Result | list[Result]):
    """Dump results with STACIE's standard output formats."""
    dn_out = Path("tests/outputs")
    dn_out.mkdir(exist_ok=True)
    plot_results(dn_out / f"{prefix}.pdf", res)
    with open(dn_out / f"{prefix}.txt", "w") as fh:
        fh.write(summarize_results(res))


def register_result(regtest, res: Result):
    """Register the result of a test with regtest.

    Parameters
    ----------
    res
        The ``stacie.estimate.Result`` to register.
    """
    with regtest:
        print(f"acint = {res.acint:.4e} ± {res.acint_std:.4e}")
        if "corrtime_exp" in res.props:
            print(f"corrtime exp = {res.corrtime_exp:.4e} ± {res.corrtime_exp_std:.4e}")
        print(f"corrtime int = {res.corrtime_int:.4e} ± {res.corrtime_int_std:.4e}")
        print("---")


@pytest.mark.parametrize(("name", "model", "fcut_max"), CASES)
@pytest.mark.parametrize("criterion", CRITERIA)
@pytest.mark.parametrize("full", [True, False])
def test_case_scan(
    regtest, name: str, model: SpectrumModel, fcut_max: float, criterion: Callable, full: bool
):
    spectrum = Spectrum(**np.load(f"tests/inputs/spectrum_{name}.npz"))
    if name == "broad":
        spectrum = spectrum.without_zero_freq()
    result = estimate_acint(spectrum, model, fcut_max=2 * fcut_max, cutoff_criterion=criterion)
    register_result(regtest, result)
    prefix = f"scan_{name}_{model.name}_{criterion.name}"
    output_test_result(prefix, result)


def test_plot_multi():
    cases = [
        ("white", ExpPolyModel([0, 2]), 0.1),
        ("broad", ExpPolyModel([0, 2]), 0.1),
    ]
    results = [
        estimate_acint(
            Spectrum(**np.load(f"tests/inputs/spectrum_{name}.npz")),
            model,
            fcut_max=fcut_max,
            cutoff_criterion=CV2LCriterion(),
        )
        for name, model, fcut_max in cases
    ]
    output_test_result("multi", results)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Unit tests for ``stacie.spectrum``."""

import numpy as np
import pytest
from numpy.testing import assert_equal

from stacie.spectrum import compute_spectrum


@pytest.mark.parametrize("use_iter", [False, True])
@pytest.mark.parametrize("prefactors_none", [False, True])
def test_basics(use_iter: bool, prefactors_none: bool):
    sequences = np.array(
        [
            [0.66134257, 1.69596962, 2.08533685, 0.62396761, -0.21445517, 1.2226847],
            [-0.66384362, -0.55499254, -1.84284631, 0.3352769, 0.86237774, 0.1605811],
        ]
    )
    prefactor = 0.34
    timestep = 10.0
    if prefactors_none:
        spectrum = compute_spectrum(
            [(prefactor, sequences[0]), (prefactor, sequences[1])]
            if use_iter
            else (prefactor, sequences),
            prefactors=None,
            timestep=timestep,
        )
    else:
        spectrum = compute_spectrum(
            iter(sequences) if use_iter else sequences,
            prefactors=[prefactor] * 2 if use_iter else prefactor,
            timestep=timestep,
        )
    # Test simple properties.
    assert spectrum.nfreq == 4
    assert_equal(spectrum.ndofs, [2, 4, 4, 2])
    assert spectrum.nstep == 6
    assert spectrum.timestep == timestep
    assert spectrum.freqs[0] == 0.0
    assert len(spectrum.freqs) == 4
    assert spectrum.freqs[1] == pytest.approx(1 / (6 * timestep))
    assert spectrum.amplitudes_ref is None
    # Test the DC-component.
    scale = 0.5 * prefactor * timestep / sequences.shape[1]
    dccomp = (sequences.sum(axis=1) ** 2).mean()
    assert spectrum.amplitudes[0] == pytest.approx(dccomp * scale)
    # Test the Plancherel theorem (taking into account RFFT conventions).
    sumsq = (sequences**2).sum()
    assert (spectrum.amplitudes * spectrum.ndofs).sum() == pytest.approx(
        sumsq * 0.5 * prefactor * timestep
    )
    # Test removing the zero frequency
    spectrum2 = spectrum.without_zero_freq()
    assert_equal(spectrum2.ndofs, spectrum.ndofs[1:])
    assert_equal(spectrum2.freqs, spectrum.freqs[1:])
    assert_equal(spectrum2.amplitudes, spectrum.amplitudes[1:])
    assert spectrum2.amplitudes_ref is None


def test_single():
    sequence = np.array([0.66134257, 1.69596962, 2.08533685, 0.62396761, -0.21445517, 1.2226847])
    prefactor = 0.25
    timestep = 2.5
    spectrum = compute_spectrum(sequence, prefactors=prefactor, timestep=timestep)
    # Test simple properties.
    assert spectrum.nfreq == 4
    assert_equal(spectrum.ndofs, [1, 2, 2, 1])
    assert spectrum.nstep == 6
    assert spectrum.timestep == timestep
    assert spectrum.freqs[0] == 0.0
    assert len(spectrum.freqs) == 4
    assert spectrum.freqs[1] == pytest.approx(1 / (6 * timestep))
    assert spectrum.amplitudes_ref is None
    # Test the DC-component.
    scale = 0.5 * prefactor * timestep / sequence.shape[0]
    dccomp = sequence.sum() ** 2
    assert spectrum.amplitudes[0] == pytest.approx(dccomp * scale)
    # Test the Plancherel theorem (taking into account RFFT conventions).
    sumsq = (sequence**2).sum()
    assert (spectrum.amplitudes * spectrum.ndofs).sum() == pytest.approx(
        sumsq * 0.5 * prefactor * timestep
    )
    # Test removing the zero frequency
    spectrum2 = spectrum.without_zero_freq()
    assert_equal(spectrum2.ndofs, spectrum.ndofs[1:])
    assert_equal(spectrum2.freqs, spectrum.freqs[1:])
    assert_equal(spectrum2.amplitudes, spectrum.amplitudes[1:])
    assert spectrum2.amplitudes_ref is None


def test_mixed():
    sequences = [
        np.array([0.66134257, 1.69596962, 2.08533685, 0.62396761, -0.21445517, 1.2226847]),
        np.array([-0.66384362, -0.55499254, -1.84284631, 0.3352769, 0.86237774, 0.1605811]),
    ]
    prefactors1 = 3.1
    prefactors2 = [3.1, 3.1]
    spectrum1 = compute_spectrum(sequences, prefactors=prefactors1, timestep=2.5)
    spectrum2 = compute_spectrum(sequences, prefactors=prefactors2, timestep=2.5)
    assert spectrum1.amplitudes == pytest.approx(spectrum2.amplitudes)


def test_variance():
    sequences = np.array(
        [
            [0.90969991, 0.09442405, 0.98258277, 0.43112388, 0.81369217, 0.5543517, 0.10178618],
            [0.78626291, 0.9948804, 0.2698581, 0.77724192, 0.11130753, 0.51409317, 0.34561125],
        ]
    )
    print(sequences.mean() ** 2)
    s1 = compute_spectrum(sequences)
    s2 = s1.without_zero_freq()
    s3 = compute_spectrum(sequences, include_zero_freq=False)
    assert s1.variance == pytest.approx((sequences**2).mean())
    assert s2.variance == pytest.approx(np.var(sequences, ddof=1))
    assert s3.variance == pytest.approx(np.var(sequences, ddof=1))



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Algorithm to estimate the autocorrelation integral."""

import attrs
import numpy as np
from numpy.typing import NDArray
from scipy.optimize import minimize, root_scalar

from .conditioning import ConditionedCost
from .cost import LowFreqCost
from .cutoff import WEIGHT_EPS, CutoffCriterion, CV2LCriterion, switch_func
from .model import SpectrumModel, guess
from .spectrum import Spectrum
from .utils import PositiveDefiniteError, UnitConfig, label_unit, mixture_stats, robust_posinv

__all__ = ("Result", "estimate_acint", "fit_model_spectrum", "summarize_results")


@attrs.define
class Result:
    """Container class holding all the results of the autocorrelation integral estimate."""

    spectrum: Spectrum = attrs.field()
    """The input spectrum from which the autocorrelation integral is estimated."""

    model: SpectrumModel = attrs.field()
    """The model used to fit the low-frequency part of the spectrum."""

    cutoff_criterion: CutoffCriterion = attrs.field()
    """The criterion used to select or weight cutoff frequencies."""

    props: dict[str] = attrs.field()
    """The properties marginalized over the ensemble of cutoff frequencies.

    Properties of this class derive their results from information in this dictionary.
    See docstring of :func:`fit_model_spectrum` for details.
    """

    history: list[dict[str]] = attrs.field()
    """History of the cutoff optimization.

    Each item is a dictionary returned by :func:`fit_model_spectrum`,
    containing the intermediate results of the fitting process.
    They are sorted from low to high cutoff frequency.
    """

    @property
    def ncut(self) -> int:
        """The number of points where the fitting weight is larger than 1/1000."""
        return self.props["ncut"]

    @property
    def fcut(self) -> int:
        """The weighted average of the cutoff frequency."""
        return self.props["fcut"]

    @property
    def neff(self) -> int:
        """The weighted average of the effective number of frequencies used in the fit."""
        return self.props["weights"].sum()

    @property
    def acint(self) -> float:
        """The autocorrelation integral."""
        return self.props["acint"]

    @property
    def acint_std(self) -> float:
        """The uncertainty of the autocorrelation integral."""
        return self.props["acint_std"]

    @property
    def corrtime_int(self) -> float:
        """The integrated correlation time."""
        return self.props["acint"] / self.spectrum.variance

    @property
    def corrtime_int_std(self) -> float:
        """The uncertainty of the integrated correlation time."""
        return self.props["acint_std"] / self.spectrum.variance

    @property
    def corrtime_exp(self) -> float:
        """The exponential correlation time."""
        if "corrtime_exp" not in self.props:
            raise ValueError("The model does not provide an exponential correlation time.")
        return self.props["corrtime_exp"]

    @property
    def corrtime_exp_std(self) -> float:
        """The uncertainty of the exponential correlation time."""
        if "corrtime_exp_std" not in self.props:
            raise ValueError("The model does not provide an exponential correlation time.")
        return self.props["corrtime_exp_std"]


def estimate_acint(
    spectrum: Spectrum,
    model: SpectrumModel,
    *,
    neff_min: int | None = None,
    neff_max: int | None = 1000,
    fcut_min: float | None = None,
    fcut_max: float | None = None,
    fcut_spacing: float = 0.5,
    switch_exponent: float = 8.0,
    cutoff_criterion: CutoffCriterion | None = None,
    rng: np.random.Generator | None = None,
    nonlinear_budget: int = 100,
    criterion_high: float = 100,
    verbose: bool = False,
    uc: UnitConfig | None = None,
) -> Result:
    r"""Estimate the integral of the autocorrelation function.

    It is recommended to leave the keyword arguments to their default values,
    except for methodological testing.

    This function fits a model to the low-frequency portion of the spectrum and
    derives an estimate of the autocorrelation (and its uncertainty) from the fit.
    It repeats this for a range of cutoff frequencies on a logarithmic grid.
    Finally, an ensemble average over all cutoffs is computed,
    by using ``-np.log`` of the cutoff criterion as weight.

    The cutoff frequency grid is logarithmically spaced,
    with the ratio between two successive cutoff frequencies given by

    .. math::

        \frac{f_{k+1}}{f_{k}} = \exp(g_\text{sp} / \beta)

    where :math:`g_\text{sp}` is ``fcut_spacing`` and :math:`\beta` is ``switch_exponent``.

    Parameters
    ----------
    spectrum
        The power spectrum and related metadata,
        used as inputs for the estimation of the autocorrelation integral.
        This object can be prepared with the function: :py:func:`stacie.spectrum.compute_spectrum`.
    model
        The model used to fit the low-frequency part of the spectrum.
    neff_min
        The minimum effective number of frequency data points to include in the fit.
        (The effective number of points is the sum of weights in the smooth cutoff.)
        If not provided, this is set to 5 times the number of model parameters as a default.
    neff_max
        The maximum number of effective points to include in the fit.
        This parameter limits the total computational cost.
        Set to None to disable this stopping criterion.
    fcut_min
        The minimum cutoff frequency to use.
        If given, this parameter can only increase the minimal cutoff derived from ``neff_min``.
    fcut_max
        If given, cutoffs beyond this maximum are not considered.
    fcut_spacing
        Dimensionless parameter that controls the spacing between cutoffs in the grid.
    switch_exponent
        Controls the sharpness of the cutoff.
        Lower values lead to a smoother cutoff, and require fewer cutoff grid points.
        Higher values sharpen the cutoff, reveal more details, but a finer cutoff grid.
    cutoff_criterion
        The criterion function that is minimized to find the best cutoff frequency and,
        consequently, the optimal number of points included in the fit.
        If not given, the default is an instance of
        :py:class:`stacie.cutoff.CV2LCriterion`.
    rng
        A random number generator for sampling guesses of the nonlinear parameters.
        If not provided, ``np.random.default_rng(42)`` is used.
        The seed is fixed by default for reproducibility.
    nonlinear_budget
        The number of samples used for the nonlinear parameters, calculated as
        ``nonlinear_budget ** num_nonlinear``.
    criterion_high
        An high increase in the cutoff criterion value,
        used to terminate the search for the cutoff frequency.
    verbose
        Set this to ``True`` to print progress information of the frequency cutoff search
        to the standard output.
    uc
        Unit configuration object used to format the screen output.
        If not provided, the default unit configuration is used.
        See :py:class:`stacie.utils.UnitConfig` for details.
        This only affects the screen output (if any) and not the results!

    Returns
    -------
    result
        The inputs, intermediate results and outputs or the algorithm.
    """
    if rng is None:
        rng = np.random.default_rng(42)
    if neff_min is None:
        neff_min = 5 * model.npar
    if cutoff_criterion is None:
        cutoff_criterion = CV2LCriterion()
    if uc is None:
        uc = UnitConfig()

    def log(props):
        neff = props["neff"]
        criterion = props["criterion"]
        line = f"{neff:9.1f}  {criterion:10.1f}    {fcut / uc.freq_unit:{uc.freq_fmt}}"
        msg = props.get("msg")
        if msg is not None:
            line += f"  ({msg})"
        print(line)

    # Determine range of cutoff frequencies
    fcut_min0 = root_scalar(
        (lambda f: switch_func(spectrum.freqs, f, switch_exponent).sum() - neff_min),
        bracket=(spectrum.freqs[1], spectrum.freqs[-1]),
    ).root
    fcut_min = fcut_min0 if fcut_min is None else max(fcut_min0, fcut_min)
    fcut_max0 = root_scalar(
        (lambda f: switch_func(spectrum.freqs[-1], f, switch_exponent) - WEIGHT_EPS),
        bracket=(spectrum.freqs[1], spectrum.freqs[-1]),
    ).root
    fcut_max = fcut_max0 if fcut_max is None else min(fcut_max0, fcut_max)

    # Iterate over the cutoff frequency_grid
    fcut_ratio = np.exp(fcut_spacing / switch_exponent)
    history = []
    best_criterion = None
    if verbose:
        print(f"CUTOFF FREQUENCY SCAN {cutoff_criterion.name}")
        print(f"     neff   criterion  {label_unit('fcut', uc.freq_unit_str):>10s}")
        print("---------  ----------  ----------")
    icut = 0
    while True:
        fcut = fcut_min * fcut_ratio**icut
        if fcut_max is not None and fcut > fcut_max:
            if verbose:
                print(f"Reached the maximum cutoff frequency ({fcut_max}).")
            break
        if fcut > spectrum.freqs[-1]:
            if verbose:
                print(f"Reached end of spectrum ({spectrum.freqs[-1]}).")
            break
        # Compute the criterion for the current cutoff frequency.
        props = fit_model_spectrum(
            spectrum,
            model,
            fcut,
            switch_exponent,
            cutoff_criterion,
            rng,
            nonlinear_budget,
        )
        if verbose:
            log(props)
        if np.isfinite(props["criterion"]):
            history.append(props)
            criterion = props["criterion"]
            if best_criterion is None or criterion < best_criterion:
                best_criterion = criterion
            elif criterion > best_criterion + criterion_high and len(history) > 10:
                if verbose:
                    print(
                        "Cutoff criterion exceeds incumbent + margin: "
                        f"{best_criterion:.1f} + {criterion_high:.1f}."
                    )
                break
        if neff_max is not None and props["neff"] > neff_max:
            if verbose:
                print(f"Reached the maximum number of effective points ({neff_max}).")
            break
        if props.get("stop", False):
            if verbose:
                print("Scan stopped by cutoff criterion.")
            break
        icut += 1

    if len(history) == 0:
        raise ValueError("The cutoff criterion could not be computed for any cutoff frequency.")

    # Weights and cutoff frequency
    criteria = np.array([props["criterion"] for props in history])
    criteria -= criteria.min()
    fcut_weights = np.exp(-criteria)
    fcut_weights /= fcut_weights.sum()
    for fcut_weight, props in zip(fcut_weights, history, strict=False):
        props["fcut_weight"] = fcut_weight
    weights = sum(
        fcut_weight * switch_func(spectrum.freqs, props["fcut"], switch_exponent)
        for fcut_weight, props in zip(fcut_weights, history, strict=True)
    )
    ncut = np.sum(weights > WEIGHT_EPS)
    freqs = spectrum.freqs[:ncut]
    weights = weights[:ncut]

    props = {
        # Fixed parameters
        "ncut": ncut,
        "weights": weights,
        "switch_exponent": switch_exponent,
        # Some properties can only be mixed directly.
        "fcut": np.dot(fcut_weights, [props["fcut"] for props in history]),
        "cost_zscore": np.dot(fcut_weights, [props["cost_zscore"] for props in history]),
        "criterion_zscore": np.dot(fcut_weights, [props["criterion_zscore"] for props in history]),
    }
    # Use fcut_weights to mix model parameters and their covariance
    props["pars"], props["pars_covar"] = mixture_stats(
        np.array([props["pars"] for props in history]),
        np.array([props["pars_covar"] for props in history]),
        fcut_weights,
    )
    # Derive other properties from the mixed parameters, instead of mixing them directly.
    # This seems to be slightly better,
    # likely because the fcut_weights are based on parameter vectors only.
    props["amplitudes_model"] = model.compute(freqs, props["pars"], deriv=1)
    _finalize_props(props, model)

    result = Result(spectrum, model, cutoff_criterion, props, history)

    if verbose:
        print()
        print(summarize_results(result, uc=uc))
    return result


def fit_model_spectrum(
    spectrum: Spectrum,
    model: SpectrumModel,
    fcut: float,
    switch_exponent: float,
    cutoff_criterion: CutoffCriterion,
    rng: np.random.Generator,
    nonlinear_budget: int,
) -> dict[str, NDArray | float]:
    """Optimize the parameter of a model for a given spectrum and cutoff frequency.

    Parameters
    ----------
    spectrum
        The spectrum object containing the input data.
    model
        The model to be fitted to the spectrum.
    fcut
        The cutoff frequency (in frequency units) used to construct the weights.
    switch_exponent
        Controls the sharpness of the cutoff.
        Lower values lead to a smoother cutoff.
        Higher values sharpen the cutoff.
    cutoff_criterion
        The criterion function that is minimized to find the optimal cutoff
        (and thus determine the number of points to include in the fit).
    rng
        A random number generator for sampling guesses of the nonlinear parameters.
    nonlinear_budget
        The number of samples to use for the nonlinear parameters is
        ``nonlinear_budget ** num_nonlinear``

    Returns
    -------
    props
        A dictionary containing various intermediate results of the cost function calculation,
        computed for the optimized parameters.
        See Notes for details.

    Notes
    -----
    The returned dictionary contains the following items if the fit succeeds:

    - ``acint``: estimate of the autocorrelation integral
    - ``acint_var``: variance of the estimate of the autocorrelation integral
    - ``acint_std``: standard error of the estimate of the autocorrelation integral
    - ``cost_value``: cost function value
    - ``cost_grad``: cost Gradient vector (if ``deriv>=1``)
    - ``cost_hess``: cost Hessian matrix (if ``deriv==2``)
    - ``cost_hess_scales``: Hessian rescaling vector, see ``robust_posinv``.
    - ``cost_hess_rescaled_evals``: Rescaled Hessian eigenvalues
    - ``cost_hess_rescaled_evecs``: Rescaled hessian eigenvectors
    - ``cost_expected``: expected value of the cost function
    - ``cost_var``: expected variance of the cost function
    - ``cost_zscore``: z-score of the cost function
    - ``switch_exponent``: exponent used to construct the cutoff
    - ``criterion``: value of the criterion whose minimizer determines the frequency cutoff
    - ``criterion_expected``: expected value of the criterion
    - ``criterion_var``: expected variance of the criterion
    - ``criterion_zscore``: z-score of the criterion
    - ``ll``: log likelihood
    - ``pars_init``: initial guess of the parameters
    - ``pars``: optimized parameters
    - ``pars_covar``: covariance matrix of the parameters

    If the model can derive the exponential correlation time,
    The following properties are also included:

    - ``corrtime_exp``: exponential correlation time, the slowest time scale in the sequences
    - ``corrtime_exp_var``: variance of the estimated exponential correlation time
    - ``corrtime_exp_std``: standard error of the estimated exponential correlation time
    - ``exp_simulation_time``: recommended simulation time based on the exponential correlation time
    - ``exp_block_time``: recommended block time based on the exponential correlation time

    The ``ExpPolyModel`` has the following additional properties:

    - ``log_acint``: the logarithm of the autocorrelation integral
    - ``log_acint_var``: variance of the logarithm of the autocorrelation integral
    - ``log_acint_std``: standard error of the logarithm of the autocorrelation integral

    If the fit fails, the following properties are set:

    - ``criterion``: infinity
    - ``msg``: error message
    """
    # Create a switching function for a smooth cutoff
    weights = switch_func(spectrum.freqs, fcut, switch_exponent)
    ncut = (weights >= WEIGHT_EPS).sum()
    freqs = spectrum.freqs[:ncut]
    ndofs = spectrum.ndofs[:ncut]
    amplitudes = spectrum.amplitudes[:ncut]
    weights = weights[:ncut]

    # Construct the initial guess for the model parameters.
    model.configure_scales(spectrum.timestep, freqs, amplitudes)
    pars_init = guess(freqs, ndofs, amplitudes, weights, model, rng, nonlinear_budget)

    # Sanity check of the initial guess
    props = {
        "fcut": fcut,
        "ncut": ncut,
        "switch_exponent": switch_exponent,
        "neff": weights.sum(),
        "pars_init": pars_init,
    }
    if not model.valid(pars_init):
        props["criterion"] = np.inf
        props["msg"] = "init: Invalid initial parameters"
        return props

    # Construct cost function and further validate initial guess
    cost = LowFreqCost(freqs, ndofs, amplitudes, weights, model)
    if not np.isfinite(cost(pars_init, deriv=0)[0]):
        props["criterion"] = np.inf
        props["msg"] = "init: Infinite cost for initial parameters"
        return props

    # Optimize the parameters
    conditioned_cost = ConditionedCost(cost, model.par_scales, 1.0)
    opt = minimize(
        conditioned_cost.funcgrad,
        conditioned_cost.to_reduced(pars_init),
        jac=True,
        hess=conditioned_cost.hess,
        bounds=model.bounds(),
        method="trust-constr",
        options={"xtol": 1e-10, "gtol": 1e-10},
    )
    pars_opt = conditioned_cost.from_reduced(opt.x)
    props["pars"] = pars_opt
    props["cost_value"], props["cost_grad"], props["cost_hess"] = cost(pars_opt, deriv=2)

    # Compute the Hessian and its properties.
    try:
        hess_scales, evals, evecs, pars_covar = robust_posinv(props["cost_hess"])
    except PositiveDefiniteError as exc:
        props["criterion"] = np.inf
        props["msg"] = f"opt: Hessian {exc.args[0]}"
        return props
    props["cost_hess_scales"] = hess_scales
    props["cost_hess_rescaled_evals"] = evals
    props["cost_hess_rescaled_evecs"] = evecs
    props["pars_covar"] = pars_covar

    # Compute the cutoff criterion
    props.update(cutoff_criterion(spectrum, model, props))

    # Compute the z-scores
    props["cost_expected"], props["cost_var"] = cost.expected(pars_opt)
    props["cost_zscore"] = (
        (props["cost_value"] - props["cost_expected"]) / np.sqrt(props["cost_var"])
        if np.isfinite(props["cost_value"])
        else np.inf
    )
    props["criterion_zscore"] = (
        (props["criterion"] - props["criterion_expected"]) / np.sqrt(props["criterion_var"])
        if np.isfinite(props["criterion"])
        else np.inf
    )

    _finalize_props(props, model)

    # Done
    return props


def _finalize_props(props: dict[str, NDArray | float], model: SpectrumModel):
    """Add remaining properties in-place."""
    model.derive_props(props)
    std_props = {}
    for key, value in props.items():
        if key.endswith("_var"):
            std_props[f"{key[:-4]}_std"] = np.sqrt(value) if value >= 0 else np.inf
    props.update(std_props)


def summarize_results(res: Result | list[Result], uc: UnitConfig | None = None):
    """Return a string summarizing the Result object."""
    if isinstance(res, Result):
        res = [res]
    if uc is None:
        uc = UnitConfig()
    texts = []
    for r in res:
        text = GENERAL_TEMPLATE.format(
            r=r,
            uc=uc,
            model=r.model.name,
            cutoff_criterion=r.cutoff_criterion.name,
            timestep=r.spectrum.timestep / uc.time_unit,
            simtime=r.spectrum.nstep * r.spectrum.timestep / uc.time_unit,
            acint=r.acint / uc.acint_unit,
            acint_std=r.acint_std / uc.acint_unit,
            corrtime_int=r.corrtime_int / uc.time_unit,
            corrtime_int_std=r.corrtime_int_std / uc.time_unit,
            npar=len(r.props["pars"]),
            maxdof=r.spectrum.ndofs.max(),
            fcut=r.fcut / uc.freq_unit,
            neff_threshold=20 * len(r.props["pars"]),
            cost_zscore=r.props["cost_zscore"],
            criterion_zscore=r.props["criterion_zscore"],
        )
        if "corrtime_exp" in r.props:
            text += EXPONENTIAL_TEMPLATE.format(
                uc=uc,
                corrtime_exp=r.props["corrtime_exp"] / uc.time_unit,
                corrtime_exp_std=r.props["corrtime_exp_std"] / uc.time_unit,
                exp_simulation_time=r.props["exp_simulation_time"] / uc.time_unit,
                exp_simulation_time_std=r.props["exp_simulation_time_std"] / uc.time_unit,
                exp_block_time=r.props["exp_block_time"] / uc.time_unit,
                exp_block_time_std=r.props["exp_block_time_std"] / uc.time_unit,
            )
        texts.append(text)
    return "\n---\n".join(texts)


GENERAL_TEMPLATE = """\
INPUT TIME SERIES
    Time step:                     {timestep:{uc.time_fmt}} {uc.time_unit_str}
    Simulation time:               {simtime:{uc.time_fmt}} {uc.time_unit_str}
    Maximum degrees of freedom:    {maxdof}

MAIN RESULTS
    Autocorrelation integral:      {acint:{uc.acint_fmt}} ± {acint_std:{uc.acint_fmt}} \
{uc.acint_unit_str}
    Integrated correlation time:   {corrtime_int:{uc.time_fmt}} ± {corrtime_int_std:{uc.time_fmt}} \
{uc.time_unit_str}

SANITY CHECKS (weighted averages over cutoff grid)
    Effective number of points:    {r.neff:.1f} (ideally > {neff_threshold:d})
    Regression cost Z-score:       {cost_zscore:.1f} (ideally < 2)
    Cutoff criterion Z-score:      {criterion_zscore:.1f} (ideally < 2)

MODEL {model} | CUTOFF CRITERION {cutoff_criterion}
    Number of parameters:          {npar}
    Average cutoff frequency:      {fcut:{uc.freq_fmt}} {uc.freq_unit_str}
"""

EXPONENTIAL_TEMPLATE = """\
    Exponential correlation time:  {corrtime_exp:{uc.time_fmt}} ± {corrtime_exp_std:{uc.time_fmt}} \
{uc.time_unit_str}

RECOMMENDED SIMULATION SETTINGS (EXPONENTIAL CORR. TIME)
    Block time:                    {exp_block_time:{uc.time_fmt}} ± \
{exp_block_time_std:{uc.time_fmt}} {uc.time_unit_str}
    Simulation time:               {exp_simulation_time:{uc.time_fmt}} ± \
{exp_simulation_time_std:{uc.time_fmt}} {uc.time_unit_str}
"""



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Tests for ``stacie.cost``."""

import numpy as np
import pytest
from conftest import check_curv, check_deriv, check_gradient, check_hessian
from scipy import stats

from stacie.cost import LowFreqCost, entropy_gamma, logpdf_gamma, varlogp_gamma
from stacie.model import PadeModel

LOGPDF_GAMMA_CASES = [
    (1.0, 1.0, 2.0),
    [1.2, 0.9, 2.2],
    [0.3, 0.001, 0.7],
    ([0.4, 1.1, 1.5], [1.0, 2.0, 3.0], [1.2, 4.3, 8.1]),
]


@pytest.mark.parametrize(("x", "alpha", "theta_ref"), LOGPDF_GAMMA_CASES)
def test_logpdf_gamma_deriv1(x, alpha, theta_ref):
    check_deriv(lambda theta, *, deriv=0: logpdf_gamma(x, alpha, theta, deriv=deriv), theta_ref)


@pytest.mark.parametrize(("x", "alpha", "theta_ref"), LOGPDF_GAMMA_CASES)
def test_logpdf_gamma_deriv2(x, alpha, theta_ref):
    check_curv(lambda theta, *, deriv=0: logpdf_gamma(x, alpha, theta, deriv=deriv), theta_ref)


@pytest.mark.parametrize(("x", "alpha", "theta_ref"), LOGPDF_GAMMA_CASES)
def test_entropy_gamma_deriv1(x, alpha, theta_ref):
    check_deriv(lambda theta, *, deriv=0: entropy_gamma(alpha, theta, deriv=deriv), theta_ref)


@pytest.mark.parametrize(("x", "alpha", "theta_ref"), LOGPDF_GAMMA_CASES)
def test_entropy_gamma_deriv2(x, alpha, theta_ref):
    check_curv(lambda theta, *, deriv=0: entropy_gamma(alpha, theta, deriv=deriv), theta_ref)


def test_entropy():
    """Check that the entropy matches the expectation value of -log(p)."""
    alpha = 2.5
    theta = 6.0
    entropy = entropy_gamma(alpha, theta)[0]
    rng = np.random.default_rng(1234)
    x = stats.gamma.rvs(alpha, scale=theta, size=10000, random_state=rng)
    logpdf1 = stats.gamma.logpdf(x, alpha, scale=theta)
    logpdf2 = logpdf_gamma(x, alpha, theta)[0]
    assert logpdf1 == pytest.approx(logpdf2, rel=1e-8)
    check = -np.mean(logpdf1)
    assert entropy == pytest.approx(check, rel=1e-2)


def test_varlogp_gamma():
    """Check the variance of the log PDF of the gamma distribution."""
    alpha = 2.5
    theta = 6.0
    varlogp = varlogp_gamma(alpha)
    rng = np.random.default_rng(1234)
    x = stats.gamma.rvs(alpha, scale=theta, size=100000, random_state=rng)
    logpdf = stats.gamma.logpdf(x, alpha, scale=theta)
    assert varlogp == pytest.approx(np.var(logpdf), rel=1e-2)


@pytest.fixture
def mycost():
    freqs = np.linspace(0, 4.0, 10)
    model = PadeModel([0, 2], [2])
    model.configure_scales(1.0, freqs, np.ones_like(freqs))
    amplitudes = np.array([1.5, 1.4, 1.1, 0.9, 0.8, 1.0, 0.9, 0.9, 0.8, 1.1])
    ndofs = np.array([5, 10, 10, 10, 10, 10, 10, 10, 10, 5])
    weights = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5])
    return LowFreqCost(freqs, ndofs, amplitudes, weights, model)


PARS_REF_PADE = [
    [1.2, 0.9, 2.2],
    [3.0, 0.5, 2.5],
    [0.1, 4.0, 2.7],
    [108.0, 77.7, 1.8],
]


def test_vectorize_pade(mycost):
    """Check that the model is vectorized."""
    results = mycost(PARS_REF_PADE, deriv=2)
    assert len(results) == 3
    nvec = len(PARS_REF_PADE)
    assert results[0].shape == (nvec,)
    assert results[1].shape == (nvec, 3)
    assert results[2].shape == (nvec, 3, 3)
    for i, one_pars_ref in enumerate(PARS_REF_PADE):
        one_results = mycost(one_pars_ref, deriv=2)
        assert one_results[0].shape == ()
        assert one_results[1].shape == (3,)
        assert one_results[2].shape == (3, 3)
        assert (one_results[0] == results[0][i]).all()
        assert (one_results[1] == results[1][i]).all()
        assert (one_results[2] == results[2][i]).all()


@pytest.mark.parametrize("pars_ref", PARS_REF_PADE)
def test_gradient_pade(mycost, pars_ref):
    check_gradient(mycost, pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_PADE)
def test_hessian_pade(mycost, pars_ref):
    check_hessian(mycost, pars_ref)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Criteria for selecting the part of the spectrum to fit to."""

import attrs
import numpy as np
from numpy.typing import NDArray
from scipy.special import polygamma

from .model import SpectrumModel
from .spectrum import Spectrum
from .utils import PositiveDefiniteError, robust_posinv

__all__ = (
    "CV2LCriterion",
    "CutoffCriterion",
    "linear_weighted_regression",
    "switch_func",
)


def switch_func(x: NDArray[float], cutoff: float, exponent: float) -> NDArray[float]:
    """Evaluate the switching function at a given points x."""
    if np.any(x < 0):
        raise ValueError("x must be non-negative")
    return 1 / (1 + (x / cutoff) ** exponent)


WEIGHT_EPS = 1e-3


@attrs.define
class CutoffCriterion:
    """Base class for cutoff criteria.

    Subclasses should implement the ``__call__`` method.
    """

    @property
    def name(self) -> str:
        """The name of the criterion."""
        raise NotImplementedError

    def __call__(
        self,
        spectrum: Spectrum,
        model: SpectrumModel,
        props: dict[str, NDArray],
    ) -> dict[str, float]:
        """Compute the criterion for the given spectrum and model.

        Parameters
        ----------
        spectrum
            The spectrum object containing the input data.
        model
            The model to be fitted to the spectrum.
        props
            The property dictionary returned by the :py:meth:`stacie.cost.LowFreqCost.props` method.

        Returns
        -------
        results
            A dictionary with at least the following fields:

            - "criterion": minus the logarithm of a likelihood of "a good fit".
            - "criterion_expected": expected value of the negative log likelihood.
            - "criterion_var": expected variance of the negative log likelihood.
            - "msg": optional message explaining failure to compute the criterion.
            - "stop": optional flag to terminate the cutoff scan early.
        """
        raise NotImplementedError


@attrs.define
class CV2LCriterion(CutoffCriterion):
    """Criterion based on the difference between fits to two halves of the spectrum."""

    fcut_factor: float = attrs.field(default=1.25, kw_only=True)
    """The scale factor to apply to the cutoff frequency.

    If 1.0, the same part of the spectrum is used as in the full non-linear regression.
    By using a larger value, the default, the criterion also tests whether
    the fitted parameters can (somewhat) extrapolate to larger frequencies,
    which reduces the risk of underfitting.
    This results in less bias on the autocorrelation integral, but slightly larger variance.
    """

    log: bool = attrs.field(default=False, kw_only=True)
    """Whether to fit a linearized model to the logarithm of the spectrum."""

    cond: float = attrs.field(default=1e6, kw_only=True)
    """The threshold for the condition number of the preconditioned covariance matrix.

    Due to the preconditioning, the condition number should be close to 1.0.
    If not, the linear dependence of the parameters is too strong, making the fit unreliable.
    In this case, "inf" is returned as the criterion.
    """

    precondition: bool = attrs.field(default=True, kw_only=True)
    """Whether to precondition the covariance eigendecomposition.

    This option is only disabled for testing. Always leave it enabled in production.
    """

    regularize: bool = attrs.field(default=True, kw_only=True)
    """Whether to regularize the linear regression.

    This option is only disabled for testing. Always leave it enabled in production.
    It will only have an impact on very ill-conditioned fits.
    """

    @property
    def name(self) -> str:
        """The name of the criterion."""
        return f"cv2l({self.fcut_factor:.0%}{',log' if self.log else ''})"

    def __call__(
        self,
        spectrum: Spectrum,
        model: SpectrumModel,
        props: dict[str, NDArray],
    ) -> dict[str, float]:
        """The disparity between fits to two different parts of the spectrum."""
        # Compute weights for the two halves and the model
        fcut = props["fcut"]
        switch_exponent = props["switch_exponent"]
        freqs = spectrum.freqs
        weights = switch_func(freqs, self.fcut_factor * fcut, switch_exponent)
        weights1 = switch_func(freqs, 0.5 * self.fcut_factor * fcut, switch_exponent)
        weights2 = weights - weights1
        ncut = (weights > WEIGHT_EPS).sum()
        if ncut == len(freqs):
            return {
                "criterion": np.inf,
                "msg": "cv2l: Insufficient data after cutoff.",
                "stop": True,
            }
        freqs = freqs[:ncut]
        weights = weights[:ncut]
        weights1 = weights1[:ncut]
        weights2 = weights2[:ncut]
        amplitudes_model = model.compute(freqs, props["pars"], deriv=1)

        # Prepare the linear problem: transform to a basis where the covariance of
        # the non-linear regression becomes the identity matrix.
        alphas = 0.5 * spectrum.ndofs[:ncut]
        design_matrix = amplitudes_model[1].T
        if self.regularize:
            design_matrix /= props["cost_hess_scales"]
            design_matrix = np.dot(design_matrix, props["cost_hess_rescaled_evecs"])
            design_matrix /= props["cost_hess_rescaled_evals"] ** 0.5

        if self.log:
            # Construct a linear regression for the residual of the logarithm of the spectrum.
            design_matrix /= amplitudes_model[0][:, None]
            expected_values = (
                np.log(spectrum.amplitudes[:ncut])
                - polygamma(0, alphas)
                - np.log(amplitudes_model[0])
            )
            evstd = np.sqrt(polygamma(1, alphas))
        else:
            # Construct a linear regression for the residual of the spectrum.
            expected_values = spectrum.amplitudes[:ncut] - amplitudes_model[0]
            evstd = amplitudes_model[0] / np.sqrt(alphas)

        # Correct the standard deviation on the expected values for the fact
        # that they are residuals with fewer degrees of freedom than the original data.
        evstd *= np.sqrt((weights.sum() - model.npar) / weights.sum())

        # Transform equations to have unit variance on the expected values.
        design_matrix = design_matrix / evstd.reshape(-1, 1)
        expected_values = expected_values / evstd

        try:
            xs, cs = linear_weighted_regression(
                design_matrix,
                expected_values,
                np.array([weights1, weights2]),
                np.array([[1.0, -1.0]]),
                ridge=1e-6 if self.regularize else 0.0,
            )
        except ValueError as exc:
            return {
                "criterion": np.inf,
                "msg": f"cv2l: {exc.args[0]}",
            }
        xd = xs[0]
        cd = cs[0, :, 0]

        # Compute the difference between the two parameter vectors in the
        # basis of the covariance matrix of the difference
        if self.precondition:
            try:
                scales, evals, evecs, _ = robust_posinv(cd)
            except PositiveDefiniteError as exc:
                return {
                    "criterion": np.inf,
                    "msg": f"cv2l: Covariance {exc.args[0]}",
                }
            if evals.max() > self.cond * evals.min():
                return {
                    "criterion": np.inf,
                    "msg": f"cv2l: Linear dependencies in basis. {evals=}",
                }
        else:
            if not (np.isfinite(cd).all()):
                return {
                    "criterion": np.inf,
                    "msg": "cv2l: Covariance matrix is not finite.",
                }
            evals, evecs = np.linalg.eigh(cd)
            if evals.min() <= 0:
                return {
                    "criterion": np.inf,
                    "msg": "cv2l: Covariance matrix is not positive definite.",
                }
            scales = np.ones(len(evals))

        # Compute the negative log likelihood of the difference in parameters.
        delta = np.dot(evecs.T, xd / scales)
        criterion = (
            0.5 * (delta**2 / evals).sum()
            + 0.5 * np.log(2 * np.pi * evals).sum()
            + np.log(scales).sum()
        )
        expected = 0.5 * np.log(2 * np.pi * np.exp(1) * evals).sum() + np.log(scales).sum()
        variance = len(evals) / 2
        if self.regularize:
            penalty = -(
                np.log(props["cost_hess_scales"]).sum()
                + 0.5 * np.log(props["cost_hess_rescaled_evals"]).sum()
            )
            criterion += penalty
            expected += penalty

        return {"criterion": criterion, "criterion_expected": expected, "criterion_var": variance}


def linear_weighted_regression(
    dm: NDArray[float],
    ev: NDArray[float],
    ws: NDArray[float],
    lc: NDArray[float] | None = None,
    ridge: float = 0.0,
) -> tuple[NDArray[float], NDArray[float]]:
    """Perform a linear regression with multiple weight vectors.

    This is a helper function for cv2l_criterion.

    Parameters
    ----------
    dm
        The design matrix.
        Shape ``(neq, npar)``, where ``neq`` is the number of equations
        and ``npar`` is the number of parameters.
    ev
        The expected values, with standard normal measurement errors.
        Shape ``(neq,)``.
    ws
        A set of weight vectors for the rows of dm (equations).
        Shape ``(nw, neq)``, where ``nw`` is the number of weight vectors.
    lc
        Linear combinations of solutions for different weights to be computed.
        Shape ``(nlc, nw)``, where ``nlc`` is the number of linear combinations.
        If None, the identity matrix is used with shape ``(nw, nw)``.

    Returns
    -------
    xs
        The regression coefficients for each weight vector.
        Shape ``(nw, npar)``.
    cs
        The covariance matrices for each combination of weight vector.
        Shape ``(nw, npar, nw, npar)``.
    """
    # Precondition the design matrix by normalizing the columns.
    column_norms = np.linalg.norm(dm, axis=0)
    if np.any(column_norms == 0):
        raise ValueError("Design matrix has zero columns.")
    column_scales = 1 / column_norms
    dm = dm * column_scales
    # Perform SVD on each weighted design matrix.
    rws = np.sqrt(ws)
    u, s, vt = np.linalg.svd(np.einsum("we,ep->wep", rws, dm), full_matrices=False)
    sinv = s / (s**2 + ridge**2)
    # Solve the problem for each weight vector.
    xs = np.einsum("wip,wi,wei,e,we->wp", vt, sinv, u, ev, rws)
    # Construct the square roots of the covariance matrix for each weight vector.
    rcs = np.einsum("wip,wi,wei,we->wpe", vt, sinv, u, rws)
    # If lc is None, use the identity matrix.
    if lc is not None:
        # Work out the requested linear combinations.
        xs = np.einsum("lw,wp->lp", lc, xs)
        rcs = np.einsum("lw,wpe->lpe", lc, rcs)
    cs = np.einsum("wpe,vqe->wpvq", rcs, rcs)
    # Account for the preconditioning to get results for the original problem.
    xs = np.einsum("wp,p->wp", xs, column_scales)
    cs = np.einsum("wpvq,p,q->wpvq", cs, column_scales, column_scales)
    return xs, cs



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""The STACIE package."""

from .cutoff import CV2LCriterion
from .estimate import estimate_acint
from .model import ExpPolyModel, LorentzModel, PadeModel
from .plot import plot_extras, plot_fitted_spectrum, plot_results, plot_spectrum
from .spectrum import compute_spectrum
from .utils import UnitConfig

__all__ = (
    "CV2LCriterion",
    "ExpPolyModel",
    "LorentzModel",
    "PadeModel",
    "UnitConfig",
    "__version__",
    "__version_tuple__",
    "compute_spectrum",
    "estimate_acint",
    "plot_extras",
    "plot_fitted_spectrum",
    "plot_results",
    "plot_spectrum",
)
